import { BehaviorSubject, Observable, Observer, Unsubscribable } from "rxjs"
import { WrappedMessage } from "../utils/message.ordering"

/* EVENT BUS */
export interface GeneralEvent<T> {
    id: string,
    type: EventType,
    event: EventMessage,
    date: Date,
    data: T
    transport: TransportType,
    transportServiceId: string | null
}


/* MANAGEMENT */
export interface MessageTransmissionManagerInterface<T> {
    getTransmissionObject(config: TransmissionConfig, role: TransmissionRole): MessageTransmissionInterface<T> | null

    // Previous version
    getTransmitter(config: TransmissionConfig): MessageTransmissionInterface<T>
    getReceiver(config: TransmissionConfig): MessageReceiverInterface<T>
}

export interface AdapterManagerInterface {
    subscribeForAdapters(): Observable<AdapterInterface<any>>
}


/* TRANSMISSION COMPONENTS */
export interface MessageTransmissionInterface<T> {
    getTransmissionProfile(type?: `config` | `id` | `role`): TransmissionProfile | TransmissionConfig | string | undefined
    subscribe(observer: Observer<T> | Observable<T>): Unsubscribable | null
}

export interface MessageReceiverInterface<T> extends MessageTransmissionInterface<T> {
    getReceivables(): Observable<any>
}

export interface MessageTransmitterInterface<T> extends MessageTransmissionInterface<T> {
    emit(message: any): void
}

export interface MessageRequestResponseInterface<T> extends MessageTransmissionInterface<T> {
    send(message: any): Observable<any>
}

/* ADAPTER COMPONENTS */
export interface AdapterInterface<T> {
    subscribe(observer: Observer<T>): Unsubscribable
    getAdapterProfile(type?: `id` | `clientId` | `role` | `transportId` | `transportType` | `connectionState`): AdapterProfile | string | Observable<ConnectionState> | undefined
}

export interface TransmitterAdapterInterface<T> extends AdapterInterface<T> {
    emit(selfId: string, message: WrappedMessage): void
}

export interface ReceiverAdapterInterface<T> extends AdapterInterface<T> {
    subscribeForIncoming(): Observable<any>
}


// /* Utils */
export interface TransportMessage {
    id: string,
    dateCreated: Date,
    transport: TransportType,
    target: string,
    source: string
    payload: any,
}

export interface FisMessage {
    header: {
        messageID: string,
        messageName: `NotificationMessage` | `ResponseMessage` | `RequestMessage`
    },
    data: any
}


export type TransportType = 'Websocket' | 'Http' | 'TCP' | undefined

export type EventType = `General Event` | 'Transport Event' | 'Transmission Event' | 'Adapter Event'
export type TransmissionRole = `Transmitter` | 'Receiver' | 'RequestResponse'
export type EventMessage = 'Server Started' | 'New Client' | 'Client Connected' | 'Client Re-connected' | 'Client Disconnected' | `Server Disconnected` | 'New Message' | `Notification` | `New Server` | `Server Connected` | `New Transport` | 'New Adapter' | 'Re-Flush' | 'New Transport' | 'New Transmission' | 'Adapter Manager Started' | 'Available Transport'
export type ConnectionState = 'ONLINE' | 'OFFLINE'

export interface TransportServiceInterface {
    getInfo(): TransportServiceProfile
    emit(message: TransportMessage): void
    subscribeForEvent(): Observable<GeneralEvent<any>> //all messages and whatever event will go through this, easier to implemnet across different transport protocol
}

export interface TransportServiceProfile {
    transportServiceId: string,
    transport: TransportType
}

export interface ClientObject {
    clientId: string,
    dateCreated: Date,
}

export interface TransportSet {
    transport: TransportType,
    port: number
}

export interface AdapterProfile {
    id: string,
    clientId: string,
    role: TransmissionRole,
    transportType: TransportType,
    transportService: TransportServiceInterface,
    connectionState: BehaviorSubject<ConnectionState>
}

export interface TransmissionProfile {
    id: string
    role: TransmissionRole
    config: TransmissionConfig
}

export interface TransmissionConfig {
    source: string
    target: string
}