/* Same as Test 1 in transmitter.ts. This one will reflect the UI environment, in that it will use is browser env.
For some cases, it is important, like websocket or grpc, or even http, because the way the transport service set up
is very different. Need to declare. If not, be default, transmission manager and adapter manager will assume they are
from the backned environment, and thus will attempt tp use dotenv configuration.

Generally the layout is more or less the same, with the exception of having to declare the isClient boolean to true.
The rest is pretty much the same. There is no message producer as in transmitter class, because I was too lazy to
separate them out. This receiver file was written after, but the concept is the same. Both transmitter and receiver
can assume dual roles. So it doesn't really matter.

Please note: For further testing, can just copy and paste from the transmitter.ts, with some tweaks on getting
profile information as well as setting the isClient boolean.*/
import { filter, interval, map, Observable, Observer, Subject, Subscription, take, Unsubscribable } from "rxjs";
import { v4 as uuidv4 } from 'uuid'
import { MessageTransmissionManager } from "../transmission/msg.transmission.manager";
import ConsoleLogger from "../utils/log.utils";
import { FisMessage, GeneralEvent, TransportType, TransportServiceInterface, TransportSet, TransmissionProfile, TransportMessage, TransmissionConfig } from "../interface/interface";
import config from '../config/config.json';
import clientProfile from '../../clients/clientprofile.json'
import { MessageTransmissionReceiver } from "../transmission/msg.transmission.receiver";
import { MessageTransmissionTransmitter } from "../transmission/msg.transmission.transmitter";
import { MessageTransmissionRequestResponse } from "../transmission/msg.transmission.request-response";
import { setUpTransportService, sortTransportFromEnv } from "../utils/transport.utils";
import { MessageProducer } from "./messageProducer";
class Supervisor {
    private console = new ConsoleLogger('Supervisor', ['base'])
    private isClient: boolean = true
    private clientIncomingMessage: Subject<FisMessage> = new Subject()
    private messageProducer!: MessageProducer
    private transmissionManager!: MessageTransmissionManager
    private event!: Subject<GeneralEvent<any>>
    private transportSet: TransportSet[] = []
    private transportServiceArray: TransportServiceInterface[] = []
    private config: TransmissionConfig = {
        target: clientProfile.clientId,
        source: clientProfile.id
    }

    constructor() {
        this.event = new Subject<GeneralEvent<any>>()
        // set up all the transportServices
        sortTransportFromEnv(this.transportSet)
        this.transportSet.forEach(set => {
            setUpTransportService(set, this.event, this.transportServiceArray, config.connection.transmitter)
        })
        this.tieInAdapterWithExistingTransportServices(this.event)
        this.messageProducer = new MessageProducer(this.clientIncomingMessage)
        this.transmissionManager = new MessageTransmissionManager(this.event, this.isClient)
        this.startMessageTransmission()
        this.clientIncomingMessage.subscribe(message => this.console.log({ message: `Received ${message.header.messageID ?? `Undefined`}` }))

    }

    // Testing here. Just comment out or uncomment out the parts you want to test
    private startMessageTransmission(): void {
        /* Actor Version <All operations are synchronized for now>  [Can set to async at a later time]*/
        let receiver: MessageTransmissionReceiver<any> = this.transmissionManager.getTransmissionObject(this.config, `Receiver`) as MessageTransmissionReceiver<any>
        let receiverSubscription = receiver.subscribe(this.clientIncomingMessage as Observer<any>) as Unsubscribable
        setTimeout(() => {
            receiverSubscription!.unsubscribe()
            this.console.log({ message: `Stop receiving messag after 10 seconds` })
        }, 20000)

        /* Primitive Version */
        // let transmitter: MessageTransmissionTransmitter<any> = this.transmissionManager.getTransmitter(this.config) as MessageTransmissionTransmitter<any>
        // let receiver: MessageTransmissionReceiver<any> = this.transmissionManager.getReceiver(this.config) as MessageTransmissionReceiver<any>
        // let requestResponse: MessageTransmissionRequestResponse<any> = this.transmissionManager.getRequestResponse(this.config, transmitter, receiver)

        // emit Message only
        // this.emitMessage(transmitter, this.messageProducer.getNotificationMessage())
        // this.emitMessage(transmitter, this.messageProducer.getOutgoingMessages())

        // receive Message only
        // this.streamMessage(receiver).subscribe((message: FisMessage) => {
        //     this.console.log({ message: `Received message ${message.header.messageID}` })
        //     this.clientIncomingMessage.next(message)
        // })

        // request-response emulation
        // this.sendMessage(requestResponse, { header: { messageID: `123`, messageName: `RequestMessage` }, data: `Test Data` } as FisMessage).subscribe({
        //     next: message => this.console.log({ message: `Received response message for request ${(message as FisMessage).header.messageID}` }),
        //     error: error => this.console.error({ message: `Something went wrong`, details: error }),
        //     complete: () => this.console.log({ message: `Request completed` })
        // })
    }

    private emitMessage(transmitter: MessageTransmissionTransmitter<any>, source: Subject<any> | Observable<any>): void {
        source.subscribe(message => transmitter.emit(message))
    }

    private sendMessage(requestResponse: MessageTransmissionRequestResponse<any>, message: any): Observable<any> {
        return new Observable((response: Observer<any>) => {
            requestResponse.send(message).subscribe({
                next: respMessage => {
                    response.next(respMessage)
                },
                error: error => {
                    this.console.error({ message: error })
                    response.error(error)
                },
                complete: () => {
                    this.console.log({ message: `Response Completed for message: ${message.id ?? 'undefined message'}` })
                    response.complete()
                }
            })
        })
    }

    // Just for clarification, this is incoming not outgoing stream. Apologies in advanced for the lack of a better term.
    private streamMessage(receiverInstance: MessageTransmissionReceiver<any>): Observable<any> {
        return new Observable((response: Observer<any>) => {
            const subscription: Subscription = receiverInstance.getReceivables().subscribe({
                next: message => {
                    response.next(message)
                },
                error: error => response.error(error),
                complete: () => response.complete()
            })

            // Clean up on unsubscription
            return () => {
                subscription.unsubscribe();
            };
        })
    }


    // this one is set up specifically for the case if there's already existing transport services ready, before the instantation of adapter manager.
    // this is so that the adapter can get all the existing available transport whilst also listening to new ones in the future
    private tieInAdapterWithExistingTransportServices(eventBus: Subject<GeneralEvent<any>>): void {
        const subscription: Subscription = eventBus.pipe(
            filter(event => event.type === `Adapter Event`),
            filter(event => event.event === `Adapter Manager Started`)
        ).subscribe(event => {
            eventBus.next({
                id: uuidv4(),
                type: `General Event`,
                event: `Available Transport`,
                date: new Date(),
                data: this.transportServiceArray,
                transport: undefined,
                transportServiceId: null
            })
        })
    }
}

let supervisor = new Supervisor()
