/* Test 1 Scenario:
So for testing, there's a few things that needs to be set up for this 1 on 1 transmission scenario to play out. First thing first, 
the 'clients' that are using this transmission is presumed. We don't know who, so we are going to pretend that they are there. Hence 
the message producer and this supervisor class that I have coded is going to fill in that supposed role. Additionally, transport services
should be set up beforehand, to ensure that there are transport protocol ready, so that this Message Transmission components can start
listening to it. 
Note: However, that doesn't mean that if there's no tranport protocol available, it will not work. The trannsmission manager and adapter
manager will just subscribe and listen to the global event for any new transport protocal and then perform the necessary subscription to 
it. */
import { filter, interval, map, Observable, Observer, Subject, Subscription, take } from "rxjs";
import { v4 as uuidv4 } from 'uuid'
import { MessageTransmissionManager } from "../transmission/msg.transmission.manager";
import ConsoleLogger from "../utils/log.utils";
import { FisMessage, GeneralEvent, TransportType, TransportServiceInterface, TransportSet, TransmissionProfile, TransportMessage, TransmissionConfig } from "../interface/interface";
import serverProfile from '../../clients/serverprofile.json'
import { MessageTransmissionReceiver } from "../transmission/msg.transmission.receiver";
import { MessageTransmissionTransmitter } from "../transmission/msg.transmission.transmitter";
import { MessageTransmissionRequestResponse } from "../transmission/msg.transmission.request-response";
import { setUpTransportService, sortTransportFromEnv } from "../utils/transport.utils";
import { MessageProducer } from "./messageProducer";
class Supervisor {
    private console = new ConsoleLogger('Supervisor', ['base'])
    private clientIncomingMessage: Subject<FisMessage> = new Subject()
    private messageProducer!: MessageProducer
    private transmissionManager!: MessageTransmissionManager
    private event!: Subject<GeneralEvent<any>>
    private transportSet: TransportSet[] = []
    private transportServiceArray: TransportServiceInterface[] = []
    private config: TransmissionConfig = {
        target: serverProfile.clientId,
        source: serverProfile.id
    }

    constructor() {
        this.event = new Subject<GeneralEvent<any>>()
        // set up all the transportServices
        sortTransportFromEnv(this.transportSet)
        this.transportSet.forEach(set => {
            setUpTransportService(set, this.event, this.transportServiceArray)
        })
        this.tieInAdapterWithExistingTransportServices(this.event)
        this.messageProducer = new MessageProducer(this.clientIncomingMessage)
        this.transmissionManager = new MessageTransmissionManager(this.event)
        this.startMessageTransmission()

    }

    // Testing here. Just comment out or uncomment out the parts you want to test
    private startMessageTransmission(): void {
        /* Actor Version <All operations are synchronized for now>  [Can set to async at a later time]*/
        let transmitter: MessageTransmissionTransmitter<any> = this.transmissionManager.getTransmissionObject(this.config, `Transmitter`) as MessageTransmissionTransmitter<any>
        let transmitterSubscription = transmitter.subscribe(this.messageProducer.getNotificationMessage())
        setTimeout(() => {
            transmitterSubscription!.unsubscribe()
            this.console.log({ message: `Stop transmitting messag after 10 seconds` })
        }, 20000)

        /* Primitive Version */
        // let transmitter: MessageTransmissionTransmitter<any> = this.transmissionManager.getTransmitter(this.config) as MessageTransmissionTransmitter<any>
        // let receiver: MessageTransmissionReceiver<any> = this.transmissionManager.getReceiver(this.config) as MessageTransmissionReceiver<any>
        // let requestResponse: MessageTransmissionRequestResponse<any> = this.transmissionManager.getRequestResponse(this.config, transmitter, receiver)

        // emit Message only
        // this.emitMessage(transmitter, this.messageProducer.getNotificationMessage())
        // this.emitMessage(transmitter, this.messageProducer.getOutgoingMessages())

        // receive Message only
        // this.streamMessage(receiver).subscribe((message: FisMessage) => {
        //     this.console.log({ message: `Received message ${message.header.messageID}` })
        //     this.clientIncomingMessage.next(message)
        // })

        // request-response emulation
        // this.sendMessage(requestResponse, { header: { messageID: `123`, messageName: `RequestMessage` }, data: `Test Data` } as FisMessage).subscribe({
        //     next: message => this.console.log({ message: `Received response message for request ${(message as FisMessage).header.messageID}` }),
        //     error: error => this.console.error({ message: `Something went wrong`, details: error }),
        //     complete: () => this.console.log({ message: `Request completed` })
        // })
    }

    private emitMessage(transmitter: MessageTransmissionTransmitter<any>, source: Subject<any> | Observable<any>): void {
        source.subscribe(message => transmitter.emit(message))
    }

    private sendMessage(requestResponse: MessageTransmissionRequestResponse<any>, message: any): Observable<any> {
        return new Observable((response: Observer<any>) => {
            requestResponse.send(message).subscribe({
                next: respMessage => {
                    response.next(respMessage)
                },
                error: error => {
                    this.console.error({ message: error })
                    response.error(error)
                },
                complete: () => {
                    this.console.log({ message: `Response Completed for message: ${message.id ?? 'undefined message'}` })
                    response.complete()
                }
            })
        })
    }

    // Just for clarification, this is incoming not outgoing stream. Apologies in advanced for the lack of a better term.
    private streamMessage(receiverInstance: MessageTransmissionReceiver<any>): Observable<any> {
        return new Observable((response: Observer<any>) => {
            const subscription: Subscription = receiverInstance.getReceivables().subscribe({
                next: message => {
                    response.next(message)
                },
                error: error => response.error(error),
                complete: () => response.complete()
            })

            // Clean up on unsubscription
            return () => {
                subscription.unsubscribe();
            };
        })
    }

    // Server to be set up as well as acquiring client information if needed. Like in the case for grpc and socket. Http not requ`ired.
    // private setUpTransportService(transportSet: TransportSet, event: Subject<GeneralEvent<any>>, isClient?: boolean): void {
    //     try {
    //         let transportService: TransportServiceInterface = this.instantiateTransportService(transportSet.transport, event)
    //         this.transportServiceArray.push(transportService)
    //         if (transportService instanceof WebsocketTransportService) {
    //             this.console.log({ message: `Just Double Checking... this is websocket` })
    //             if (isClient) {
    //                 // please note this is subject to change depending on the UI environemnt. Angular has their own built in function to read json file based on Swopt-UI
    //                 transportService.startClient(config.connection.transmitter)
    //             } else {
    //                 transportService.startServer(transportSet.port);
    //             }
    //         } else if (transportService instanceof HttpTransportService) {
    //             this.console.log({ message: `Just Double Checking... this is http` })
    //             // Additional Http-specific setup if needed.
    //             if (isClient) {
    //                 transportService.startClient(config.connection.transmitter)
    //             } else {
    //                 transportService.startServer(transportSet.port)
    //             }
    //         }
    //         // additional transport protocol here...
    //     } catch (error) {
    //         this.console.error({ message: 'Fail to set transport. Error in setting up transport', details: error })
    //     }
    // }

    // private instantiateTransportService(transportType: TransportType, event: Subject<GeneralEvent<any>>): TransportServiceInterface {
    //     if (transportType === 'Websocket') {
    //         return new WebsocketTransportService(event)
    //     }
    //     else if (transportType === 'Http') {
    //         return new HttpTransportService(event)
    //     } else {
    //         throw new Error(`No Transport Service Instantiated`)
    //     }
    //     // additional transport protocol here...
    // }

    // // get the transpor information from dotenv file. Plase note this is just testing only. Real production may have their own way of defining their transport protocl. Please adapte accordingly.
    // private sortTransportFromEnv(transportSet: TransportSet[]): void {
    //     let transportList: string[] = process.env.Transport?.split(',') || []
    //     let portList: number[] = (process.env.PORT?.split(',') || []).map(port => Number(port));
    //     transportList.forEach((transport, index) => {
    //         transportSet.push({ transport: transport, port: portList[index] } as unknown as TransportSet)
    //     })
    //     this.console.log({ message: 'TransportSetList', details: this.transportSet })
    // }

    // this one is set up specifically for the case if there's already existing transport services ready, before the instantation of adapter manager.
    // this is so that the adapter can get all the existing available transport whilst also listening to new ones in the future
    private tieInAdapterWithExistingTransportServices(eventBus: Subject<GeneralEvent<any>>): void {
        const subscription: Subscription = eventBus.pipe(
            filter(event => event.type === `Adapter Event`),
            filter(event => event.event === `Adapter Manager Started`)
        ).subscribe(event => {
            eventBus.next({
                id: uuidv4(),
                type: `General Event`,
                event: `Available Transport`,
                date: new Date(),
                data: this.transportServiceArray,
                transport: undefined,
                transportServiceId: null
            })
        })
    }
}



let supervisor = new Supervisor()
