/* General Note: Mesasge transmission manager instantiate transmission component, but it itself is not concerned with the state of the 
client. So, whether or not if the client is online, or what-ever transport protocol they're using is none of this manager's concern. */
import { MessageTransmissionTransmitter } from "./msg.transmission.transmitter";
import { MessageTransmissionReceiver } from "./msg.transmission.receiver";
import { AdapterManager } from "../adapters/adapter.manager";
import { v4 as uuidv4 } from 'uuid'
import { MessageTransmissionRequestResponse } from "./msg.transmission.request-response";
import { filter, map, Subject } from "rxjs";
import ConsoleLogger from "../utils/log.utils";
import { AdapterInterface, AdapterManagerInterface, GeneralEvent, MessageReceiverInterface, MessageTransmissionInterface, MessageTransmitterInterface, TransmissionConfig, TransmissionProfile, TransmissionRole } from "../interface/interface";
import { MessageTransmissionManagerBase } from "../base/msg.transmission.manager.base";

export class MessageTransmissionManager extends MessageTransmissionManagerBase<any> {

    constructor(event: Subject<GeneralEvent<any>>, browserEnv?: boolean, adapterManager?: AdapterManager) {
        super()
        this.console = new ConsoleLogger(`MessageTransmissionManager`, ['managers'])
        if (browserEnv) this.browserEnv = browserEnv
        this.console.log({ message: `Contructing self... ${this.browserEnv ? `is receiving end` : `is not browser env`}` })
        this.globalEvent = event
        this.transmissionEvent = new Subject<any>()
        this.adapterManager = adapterManager ? adapterManager : new AdapterManager(event, this.browserEnv)
        this.setupTransmissionEvent(this.adapterManager)
    }

    public getTransmissionObject(config: TransmissionConfig, role: TransmissionRole): MessageTransmissionInterface<any> | null {
        if (role === `Transmitter`) return this.getTransmitter(config)
        if (role === `Receiver`) return this.getReceiver(config)
        else {
            let transmitter = this.transmissionObjects.find(obj => (obj.getTransmissionProfile() as TransmissionProfile).config.target === config.target && (obj.getTransmissionProfile() as TransmissionProfile).role === 'Transmitter')
            let receiver = this.transmissionObjects.find(obj => (obj.getTransmissionProfile() as TransmissionProfile).config.target === config.target && (obj.getTransmissionProfile() as TransmissionProfile).role === 'Receiver')
            if (transmitter && receiver) {
                return this.getRequestResponse(config, transmitter as MessageTransmitterInterface<any>, receiver as MessageReceiverInterface<any>)
            } else {
                return null
            }
        }
    }


    public getTransmitter(config: TransmissionConfig): MessageTransmitterInterface<any> {
        let transmitterInstance = new MessageTransmissionTransmitter(config, this.transmissionEvent)
        this.transmissionObjects.push(transmitterInstance)
        return transmitterInstance
    }

    public getReceiver(config: TransmissionConfig): MessageReceiverInterface<any> {
        let receiverInstance = new MessageTransmissionReceiver(config, this.transmissionEvent)
        this.transmissionObjects.push(receiverInstance)
        return receiverInstance
    }

    public getRequestResponse(config: TransmissionConfig, transmitterInstance: MessageTransmitterInterface<any>, receiverInstance: MessageReceiverInterface<any>): MessageTransmissionRequestResponse<any> {
        return new MessageTransmissionRequestResponse(config, transmitterInstance, receiverInstance)
    }

    // Subscribe for new adapters from Adapter Manager
    private setupTransmissionEvent(adapterManager: AdapterManagerInterface): void {
        adapterManager.subscribeForAdapters().pipe(
            map((adapter: AdapterInterface<any>) => {
                return {
                    id: uuidv4(),
                    type: `Transmission Event`,
                    event: `New Adapter`,
                    date: new Date(),
                    data: adapter,
                    transport: adapter.getAdapterProfile(`transportType`)
                } as GeneralEvent<AdapterInterface<any>>
            })
        ).subscribe(this.transmissionEvent)
    }

}


